'use strict';

Object.defineProperty(exports, '__esModule', {
  value: true
});

var _interopRequireDefault = function (obj) { return obj && obj.__esModule ? obj : { 'default': obj }; };

var _toConsumableArray = function (arr) { if (Array.isArray(arr)) { for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i]; return arr2; } else { return Array.from(arr); } };

var _relative$basename$join$dirname = require('path');

var _fs = require('fs');

var _fs2 = _interopRequireDefault(_fs);

'use babel';

var meta = {};

exports['default'] = function (babel) {
  var Transformer = babel.Transformer;
  var t = babel.types;

  var skipper = {
    enter: function enter(node) {
      if (!skip()) {
        meta.moduleExports.push(node);
      }

      enterSkip(node);
    },

    exit: function exit(node) {
      exitSkip(node);
    }
  };

  return new Transformer('aurelia-babel-plugin', {
    Program: {
      enter: function enter(node, parent) {
        meta = {};
        var _state$opts = this.state.opts;
        var filename = _state$opts.filename;
        var moduleRoot = _state$opts.moduleRoot;
        var _state$opts$extra = _state$opts.extra;
        var packageName = _state$opts$extra.packageName;
        var typings = _state$opts$extra.typings;

        var moduleId = packageName + '/' + _relative$basename$join$dirname.relative(moduleRoot, filename).replace('.js', '');
        meta.root = packageName;
        meta.moduleId = moduleId;
        meta.moduleExports = [];
        meta.moduleImports = [];
        meta.skipStack = [];
        meta.outpath = _relative$basename$join$dirname.join(typings, moduleId + '.d.ts');
      },

      exit: function exit(node, parent) {
        var code = generate(meta);
        //console.log(code);
        ensureDir(meta.outpath);
        _fs2['default'].writeFileSync(meta.outpath, code);
        meta = {};
      }
    },

    ExportAllDeclaration: function ExportAllDeclaration(node) {
      meta.moduleExports.push(node);
    },

    ExportNamedDeclaration: skipper,

    ExportDefaultDeclaration: skipper,

    ClassDeclaration: skipper,

    ImportDeclaration: function ImportDeclaration(node) {
      meta.moduleImports.push(node);
    }
  });

  function skip() {
    return meta.skipStack.length > 0;
  }

  function enterSkip(node) {
    meta.skipStack.push(node);
  }

  function exitSkip(node) {
    if (node !== meta.skipStack.pop()) {
      throw new Error('skipStack unballance');
    }
  }
};

function generate(data) {
  var moduleId = data.moduleId;
  var moduleExports = data.moduleExports;
  var moduleImports = data.moduleImports;

  var str = 'declare module \'' + moduleId + '\' {\n';
  var _iteratorNormalCompletion = true;
  var _didIteratorError = false;
  var _iteratorError = undefined;

  try {
    for (var _iterator = moduleImports[Symbol.iterator](), _step; !(_iteratorNormalCompletion = (_step = _iterator.next()).done); _iteratorNormalCompletion = true) {
      var i = _step.value;

      var importStr = generateDts(i);
      if (importStr) {
        importStr = importStr.split('\n').map(function (s) {
          return '  ' + s;
        }).join('\n');
        str += importStr + '\n';
      }
    }
  } catch (err) {
    _didIteratorError = true;
    _iteratorError = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion && _iterator['return']) {
        _iterator['return']();
      }
    } finally {
      if (_didIteratorError) {
        throw _iteratorError;
      }
    }
  }

  var _iteratorNormalCompletion2 = true;
  var _didIteratorError2 = false;
  var _iteratorError2 = undefined;

  try {
    for (var _iterator2 = moduleExports[Symbol.iterator](), _step2; !(_iteratorNormalCompletion2 = (_step2 = _iterator2.next()).done); _iteratorNormalCompletion2 = true) {
      var e = _step2.value;

      var exportStr = generateDts(e);
      if (exportStr) {
        exportStr = exportStr.split('\n').map(function (s) {
          return '  ' + s;
        }).join('\n');
        str += exportStr + '\n';
      }
    }
  } catch (err) {
    _didIteratorError2 = true;
    _iteratorError2 = err;
  } finally {
    try {
      if (!_iteratorNormalCompletion2 && _iterator2['return']) {
        _iterator2['return']();
      }
    } finally {
      if (_didIteratorError2) {
        throw _iteratorError2;
      }
    }
  }

  str += '}';
  return str;
}

var exportGenerators = {
  ExportAllDeclaration: function ExportAllDeclaration(node) {
    return 'export * from \'' + getSource(node.source) + '\';';
  },

  ExportNamedDeclaration: function ExportNamedDeclaration(node) {
    if (node.declaration) {
      var classString = generateDts(node.declaration);
      if (classString) {
        return 'export ' + classString;
      } else {
        console.log('Unsupported node type ' + node.declaration.type);
      }
    } else if (node.source != null) {
      var objectExports = node.specifiers.filter(function (s) {
        return s.type === 'ExportSpecifier';
      }).map(function (s) {
        var importName = s.exported.name;
        var localName = s.local.name;

        if (!localName || localName === importName) return importName;else return '' + importName + ' as ' + localName;
      }).join(', ');

      return 'export { ' + objectExports + ' } from \'' + getSource(node.source) + '\';';
    }
  },

  ExportDefaultDeclaration: function ExportDefaultDeclaration(node) {
    var value = generateDts(node.declaration);
    if (!value) {
      console.log('Can\'t generate ' + node.declaration.type + '.');
    }

    return 'export default ' + value + ';';
  },

  ImportDeclaration: function ImportDeclaration(node) {
    var objectImports = node.specifiers.filter(function (s) {
      return s.type === 'ImportSpecifier';
    }).map(function (s) {
      var importName = s.imported.name;
      var localName = s.local.name;

      if (!localName || localName === importName) return importName;else return '' + importName + ' as ' + localName;
    }).join(', ');
    var defaultImports = node.specifiers.filter(function (s) {
      return s.type === 'ImportDefaultSpecifier';
    }).map(function (s) {
      var localName = s.local.name;

      return localName;
    });
    var namespaceImports = node.specifiers.filter(function (s) {
      return s.type === 'ImportNamespaceSpecifier';
    }).map(function (s) {
      var localName = s.local.name;

      return '* as ' + localName;
    });

    var parts = [].concat(_toConsumableArray(defaultImports), _toConsumableArray(namespaceImports));
    if (objectImports.length) {
      parts.push('{ ' + objectImports + ' } ');
    }

    return 'import ' + parts.join(', ') + ' from \'' + getSource(node.source) + '\';';
  },

  VariableDeclaration: function VariableDeclaration(node) {
    var kind = node.kind;
    var declarations = node.declarations.map(function (d) {
      var name = d.id.name;
      var typeAnnotation = d.typeAnnotation;

      var type = 'any';
      if (typeAnnotation) {
        type = getTypeAnnotation(typeAnnotation);
      }

      return '' + name + ': ' + type;
    }).join(', ');

    return '' + kind + ' ' + declarations + ';';
  },

  ClassDeclaration: function ClassDeclaration(node) {
    var name = node.id.name;
    var superClass = node.superClass;
    var body = node.body;

    var str = 'class ' + name + ' ';
    if (superClass) {
      str += 'extends ' + superClass.name + ' ';
    }
    str += '{\n';

    var _iteratorNormalCompletion3 = true;
    var _didIteratorError3 = false;
    var _iteratorError3 = undefined;

    try {
      for (var _iterator3 = body.body[Symbol.iterator](), _step3; !(_iteratorNormalCompletion3 = (_step3 = _iterator3.next()).done); _iteratorNormalCompletion3 = true) {
        var member = _step3.value;

        var memberStr = generateDts(member);
        if (memberStr) {
          memberStr = memberStr.split('\n').map(function (s) {
            return '  ' + s;
          }).join('\n');
          str += memberStr + '\n';
        } else {
          console.log('missing member type: ', member.type);
        }
      }
    } catch (err) {
      _didIteratorError3 = true;
      _iteratorError3 = err;
    } finally {
      try {
        if (!_iteratorNormalCompletion3 && _iterator3['return']) {
          _iterator3['return']();
        }
      } finally {
        if (_didIteratorError3) {
          throw _iteratorError3;
        }
      }
    }

    return str + '}';
  },

  MethodDefinition: function MethodDefinition(node) {
    var args = getArgs(node.value).join(', ');
    if (node.kind === 'constructor') {
      return 'constructor(' + args + ');';
    } else {
      var _name = node.key.name;
      var _returnType = node.value.returnType;

      var _type = 'any';
      var prefix = '';

      if (_returnType) {
        _type = getTypeAnnotation(_returnType);
      }

      if (node['static']) {
        prefix = 'static ';
      }

      return '' + prefix + '' + _name + '(' + args + '): ' + _type + ';';
    }
  },

  ClassProperty: function ClassProperty(node) {
    var name = node.key.name;
    var typeAnnotation = node.typeAnnotation;

    var type = 'any';
    var prefix = '';

    if (node['static']) {
      prefix = 'static ';
    }

    if (typeAnnotation) {
      type = getTypeAnnotation(typeAnnotation);
    }

    return '' + prefix + '' + name + ': ' + type + ';';
  },

  NewExpression: function NewExpression(node) {
    var args = generateArgs(node.arguments);
    return 'new ' + node.callee.name + '(' + args + ')';
  },

  FunctionDeclaration: function FunctionDeclaration(node) {
    var name = node.id.name;
    var returnType = node.returnType;

    var args = getArgs(node).join(', ');
    var type = 'any';

    if (returnType) {
      type = getTypeAnnotation(returnType);
    }

    return 'function ' + name + '(' + args + '): ' + type + ';';
  },

  Literal: function Literal(node) {
    return node.raw;
  }
};

function generateDts(node) {
  var fn = exportGenerators[node.type];
  if (fn) {
    return fn(node);
  }
}

function getSource(node) {
  var value = node.value;

  if (value.substring(0, 2) === './') {
    value = meta.root + value.substring(1);
  }
  return value;
}

function generateArgs(args) {
  return args.map(function (arg) {
    var str = generateDts(arg);
    if (str) return str;

    return '"Unsupported argument type ' + arg.type + '"';
  }).join(', ');
}

function getArg(p) {
  switch (p.type) {
    case 'Identifier':
      var name = p.name,
          typeAnnotation = p.typeAnnotation;

      var type = 'any';
      if (typeAnnotation) {
        type = getTypeAnnotation(typeAnnotation);
      }

      return '' + name + ': ' + type;

    case 'AssignmentPattern':
      var _p$left = p.left,
          name = _p$left.name,
          typeAnnotation = _p$left.typeAnnotation;

      var type = 'any';
      if (typeAnnotation) {
        type = getTypeAnnotation(typeAnnotation);
      }

      return '' + name + '?: ' + type;

    case 'RestElement':
      var _p$argument = p.argument,
          name = _p$argument.name,
          typeAnnotation = _p$argument.typeAnnotation;

      var type = 'any[]';
      if (typeAnnotation) {
        type = getTypeAnnotation(typeAnnotation);
      }

      return '...' + name + ': ' + type;

    default:
      console.log(p);
      throw new Error('Parameter type ' + p.type + ' not implemented');
  }
}

function getArgs(node) {
  var params = node.params;

  return params.map(getArg);
}

function getTypeAnnotation(annotated) {
  return getTypeAnnotationString(annotated.typeAnnotation);
}

function getTypeAnnotationString(annotation) {
  switch (annotation.type) {
    case 'GenericTypeAnnotation':
      var name = annotation.id.name;

      if (annotation.typeParameters) {
        var typeParameters = annotation.typeParameters.params.map(getTypeAnnotationString).join(', ');
        return '' + name + '<' + typeParameters + '>';
      }

      return name;

    case 'VoidTypeAnnotation':
      return 'void';

    case 'NumberTypeAnnotation':
      return 'number';

    case 'StringTypeAnnotation':
      return 'string';

    case 'AnyTypeAnnotation':
      return 'any';

    case 'BooleanTypeAnnotation':
      return 'boolean';

    case 'UnionTypeAnnotation':
      return annotation.types.map(getTypeAnnotationString).join(' | ');

    case 'FunctionTypeAnnotation':
      var params = annotation.params.map(getFunctionTypeAnnotationParameter).join(', ');
      var returnType = getTypeAnnotationString(annotation.returnType);
      return '(' + params + ') => ' + returnType;

    case 'ArrayTypeAnnotation':
      var elementType = getTypeAnnotationString(annotation.elementType);
      return '' + elementType + '[]';

    default:
      throw new Error('Unsupported type annotation type: ' + annotation.type);
  }
}

function getFunctionTypeAnnotationParameter(node) {
  var name = node.name.name;
  var typeAnnotation = node.typeAnnotation;
  var optional = node.optional;

  var type = 'any';
  if (typeAnnotation) {
    type = getTypeAnnotationString(typeAnnotation);
  }
  if (optional) {
    name = '' + name + '?';
  }

  return '' + name + ': ' + type;
}

function ensureDir(p) {
  var dn = _relative$basename$join$dirname.dirname(p);
  if (!_fs2['default'].existsSync(dn)) {
    ensureDir(dn);
    try {
      _fs2['default'].mkdirSync(dn);
    } catch (e) {}
  }
}
module.exports = exports['default'];